from keras.optimizers import Adam, RMSprop
from keras.layers import Lambda, Dense, Input
from keras.models import Model
import tensorflow as tf
from abc import abstractmethod
import keras.backend as K
import tensorflow_probability as tfp
tfd = tfp.distributions


class VariationalAutoEncoder:
    def __init__(self, input_shape, latent_dim):
        self.input_shape = input_shape
        self.latent_dim = latent_dim
        self.BATCH_SIZE = input_shape[0]

    @abstractmethod
    def _encoder(self):
        pass

    @abstractmethod
    def _decoder(self):
        pass

    def _sampling(self, args):
        z_mean, z_log_sigma = args
        epsilon = K.random_normal(
            shape=(self.BATCH_SIZE, self.latent_dim), mean=0.)
        return z_mean + K.exp(z_log_sigma) * epsilon

    def _get_loss(self, loss_fn):
        if loss_fn == None:
            return self._bernoulli
        elif loss_fn == "binary":
            return self._binary
        elif loss_fn == "normal":
            return self._normal
        elif loss_fn == "normalDiag":
            return self._normalDiag

    def get_compiled_model(self, *args):
        loss_fn, lr, decay, self.observation_noise, beta = args
        input_img = Input(batch_shape=self.input_shape)

        encoder = self._encoder()
        encoded = encoder(input_img)

        # Reparametrization
        self.z_mean = Dense(self.latent_dim)(encoded)
        self.z_log_sigma = Dense(self.latent_dim)(encoded)
        z = Lambda(self._sampling, name="latent")([self.z_mean, self.z_log_sigma])

        decoder = self._decoder()
        reco = decoder(z)

        self.beta = beta
        self.loss_fn = self._get_loss(loss_fn)

        # Generator model
        decoder_input = Input(shape=(self.latent_dim,))
        _reco = decoder(decoder_input)
        generator = Model(decoder_input, _reco)

        model = Model(input_img, outputs=[reco, z])
        model.compile(optimizer=RMSprop(lr=lr, decay=decay),
                      loss=self.loss_fn, metrics=[self.KL_divergence])

        return model, generator

    """
      Making it custom metric to be able to feed it to Keras API - actually no need for y_true, y_pred
    """

    def KL_divergence(self, y_true, y_pred):
        return - self.beta * 0.5 * K.mean(
            1 + self.z_log_sigma - K.square(self.z_mean) - K.exp(self.z_log_sigma), axis=-1)

    """
      For binarized input with KL term (!)
    """

    def _binary(self, x_true, x_reco):
        return -tf.nn.sigmoid_cross_entropy_with_logits(
            labels=x_true, logits=x_reco) + self.KL_divergence(None, None)

    def _bernoulli(self, x_true, x_reco):
        return -tf.reduce_mean(tfd.Bernoulli(x_reco)._log_prob(x_true)
                               ) + self.KL_divergence(None, None)

    """
      For non binarized input with KL term(!)
    """

    def _normal(self, x_true, x_reco):
        return -tf.reduce_mean(
            tfd.Normal(x_reco, scale=self.observation_noise)._log_prob(x_true)
        ) + self.KL_divergence(None, None)

    def _normalDiag(self, x_true, x_reco):
        return -tf.reduce_mean(
            tfd.MultivariateNormalDiag(
                x_reco, scale_identity_multiplier=self.observation_noise)._log_prob(x_true)) + self.KL_divergence(None, None)



from __future__ import absolute_import

from .autoencoders import DenseAutoEncoder, ConvolutionalAutoEncoder
from .csnl_util import DataGenerator
from .train import ModelTrainer
from .variational_autoencoders import DenseVAE, SmallDenseVAE, ConvolutionalVAE, SmallConvolutionalVAE
from .visualize import VAEPlotter



import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import numpy as np


class VAEPlotter:
    def __init__(self, fitted_model, datagen):
        self.model = fitted_model.model
        self.generator_model = fitted_model.generator
        self._latent_dim = fitted_model.latent_dim
        self.datagen = datagen
        self.train_images, _ = next(self.datagen.flow(self._latent_dim))
        self.test_images, _ = self.datagen.validation_data(self._latent_dim)
        self.batch_size = self.train_images.shape[0]
        self.input_shape = self.model.layers[0].output_shape
        self.image_shape = self.train_images.shape[1:]
        if self.image_shape[2] == 1:
            self.image_shape = self.image_shape[:-1]

    def grid_plot(self):
        reco_train, _ = self.model.predict(self.train_images.reshape(
            self.batch_size, *self.input_shape[1:]), batch_size=self.batch_size)

        reco_test, _ = self.model.predict(self.test_images[:self.batch_size].reshape(
            self.batch_size, *self.input_shape[1:]), batch_size=self.batch_size)

        reco = [reco_train, reco_test]
        images = [self.train_images, self.test_images]

        fig = plt.figure(figsize=(14, 7))
        outer = gridspec.GridSpec(1, 2, wspace=0.1, hspace=0.1)

        for outer_ind in range(2):
            inner = gridspec.GridSpecFromSubplotSpec(4, 4,
                                                     subplot_spec=outer[outer_ind], wspace=0.095, hspace=0.095)
            for inner_ind in range(16):
                ax = plt.Subplot(fig, inner[inner_ind])
                if inner_ind % 2 == 0:
                    if inner_ind < 4:
                        ax.set_title('Reconstructed')
                    ax.imshow(reco[outer_ind][inner_ind].reshape(
                        *self.image_shape), interpolation=None)
                else:
                    if inner_ind < 4:
                        ax.set_title(' <- Original')
                    ax.imshow(images[outer_ind][inner_ind -
                                                1].reshape(*self.image_shape), interpolation=None)
                ax.set_xticks([])
                ax.set_yticks([])
                fig.add_subplot(ax)

        fig.suptitle("Train and test reconstructions\n\n")
        plt.show()

    def generate_samples(self):
        latent_inputs = np.random.normal(size=(14*14, self._latent_dim))

        self._plot_samples(latent_inputs)

    def visualize_latent_param(self, axis=0):
        sweep = np.linspace(-5, 5, 14*14)
        latent_inputs = np.array(np.random.normal(
            size=(1, self._latent_dim)).tolist() * (14*14))
        latent_inputs = latent_inputs.reshape(14*14, self._latent_dim)
        latent_inputs[:, axis] = sweep

        self._plot_samples(latent_inputs)

    def _plot_samples(self, latent_inputs):
        recos = self.generator_model.predict(
            latent_inputs[:14*14].reshape(14*14, self._latent_dim))
        # Output 196 images
        fig, axes = plt.subplots(
            14, 14, sharex=True, sharey=True, figsize=(11, 11))

        for ind, ax in enumerate(axes.flatten()):
            ax.imshow(recos[ind].reshape(28, 28))
            ax.set_xticks([])
            ax.set_yticks([])

        fig.suptitle("Generated samples on %d - dimensional grid" %
                     self._latent_dim)
        plt.show()



from setuptools import setup
from setuptools import find_packages

setup(
    name='csnl-vae-olaralex',
    version='1.37dev',
    author="Alex Olar",
    author_email="olaralex@caesar.elte.hu",
    packages=find_packages(),
    license='Creative Commons Attribution-Noncommercial-Share Alike license',
    long_description='Variational Autoencoders with different architectures and test files',
)


from keras.preprocessing.image import ImageDataGenerator
import numpy as np
from .data_loader import DataLoader


class DataGenerator:
    def __init__(self, image_shape, batch_size, file_path='/data/scramtex_700_28px.pkl', binarize=False):
        self.DATA_LOADER = DataLoader(file_path, binarize)
        self.IMAGE_SHAPE = image_shape
        self.BATCH_SIZE = batch_size

        self.DATA_GENERATOR = ImageDataGenerator()

        self.TRAIN, self.TEST = self.DATA_LOADER.train_test_split()

    def flow(self, latent_dim):
        def train_generator(_it):
            while True:
                batch_x, batch_y = next(_it)
                val = np.ones(shape=(self.BATCH_SIZE, latent_dim))
                yield batch_x, {"target" : batch_y, "latent" : val}
        return train_generator(self.DATA_GENERATOR.flow(self.TRAIN, self.TRAIN, batch_size=self.BATCH_SIZE))

    def contrast_flow(self, latent_dim):
        def train_generator(_it):
            while True:
                batch_x, target_dict = next(_it)
                batch_y = target_dict["target"]
                val = target_dict["latent"]
                contrast = np.random.rand()
                _batch_x, _batch_y = contrast*batch_x, contrast*batch_y
                _batch_x, _batch_y = np.clip(_batch_x, 0, 1), np.clip(_batch_y, 0, 1)
                yield _batch_x, {"target" : _batch_y, "latent" : val}
        return train_generator(self.flow(latent_dim))

    def flattened_flow(self, latent_dim):
        def train_generator(_it):
            image_dim = np.prod(self.IMAGE_SHAPE)
            while True:
                batch_x, (batch_y, val) = next(_it)
                yield batch_x.reshape(batch_x.shape[0], image_dim), batch_y.reshape(batch_y.shape[0], image_dim), val
        return train_generator(self.flow(latent_dim))

    def validation_data(self, latent_dim):
        return self.TEST, {"target" : self.TEST, "latent" : np.ones(shape=(self.TEST.shape[0], latent_dim))}

    def flattened_validation_data(self, latent_dim):
        image_dim = np.prod(self.IMAGE_SHAPE)
        reshaped_test = self.TEST.reshape(self.TEST.shape[0], image_dim)
        return reshaped_test, reshaped_test, np.ones(shape=(self.BATCH_SIZE, latent_dim))
